# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals

"""
This module provides classes to perform topological analyses of structures.
"""

__author__ = "Shyue Ping Ong, Geoffroy Hautier, Sai Jayaraman"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__status__ = "Production"
__date__ = "Sep 23, 2011"

import math
from math import pi
import numpy as np
import itertools
import collections

from warnings import warn
from scipy.spatial import Voronoi
from pymatgen import PeriodicSite
from pymatgen import Element, Specie, Composition
from pymatgen.util.num_utils import abs_cap


class VoronoiCoordFinder(object):
    """
    Uses a Voronoi algorithm to determine the coordination for each site in a
    structure.

    Args:
        structure (Structure): Input structure
        target ([Element/Specie]): A list of target species to determine
            coordination for.
        cutoff (float): Radius in Angstrom cutoff to look for coordinating
            atoms. Defaults to 10.0.
    """

    def __init__(self, structure, target=None, cutoff=10.0):
        self._structure = structure
        self.cutoff = cutoff
        if target is None:
            self._target = structure.composition.elements
        else:
            self._target = target

    def get_voronoi_polyhedra(self, n):
        """
        Gives a weighted polyhedra around a site. This uses the voronoi
        construction with solid angle weights.
        See ref: A Proposed Rigorous Definition of Coordination Number,
        M. O'Keeffe, Acta Cryst. (1979). A35, 772-775

        Args:
            n (int): Site index

        Returns:
            A dict of sites sharing a common Voronoi facet with the site
            n and their solid angle weights
        """
        localtarget = self._target
        center = self._structure[n]
        neighbors = self._structure.get_sites_in_sphere(
            center.coords, self.cutoff)
        neighbors = [i[0] for i in sorted(neighbors, key=lambda s: s[1])]
        qvoronoi_input = [s.coords for s in neighbors]
        voro = Voronoi(qvoronoi_input)
        all_vertices = voro.vertices

        results = {}
        for nn, vind in voro.ridge_dict.items():
            if 0 in nn:
                if -1 in vind:
                    raise RuntimeError("This structure is pathological,"
                                       " infinite vertex in the voronoi "
                                       "construction")

                facets = [all_vertices[i] for i in vind]
                results[neighbors[sorted(nn)[1]]] = solid_angle(
                    center.coords, facets)

        maxangle = max(results.values())

        resultweighted = {}
        for nn, angle in results.items():
            # is nn site is ordered use "nn.specie" to get species, else use "nn.species_and_occu" to get species
            if nn.is_ordered:
                if nn.specie in localtarget:
                    resultweighted[nn] = angle / maxangle
            else:  # is nn site is disordered
                for disordered_sp in nn.species_and_occu.keys():
                    if disordered_sp in localtarget:
                        resultweighted[nn] = angle / maxangle

        return resultweighted

    def get_coordination_number(self, n):
        """
        Returns the coordination number of site with index n.

        Args:
            n (int): Site index
        """
        return sum(self.get_voronoi_polyhedra(n).values())

    def get_coordinated_sites(self, n, tol=0, target=None):
        """
        Returns the sites that are in the coordination radius of site with
        index n.

        Args:
            n (int): Site index.
            tol (float): Weight tolerance to determine if a particular pair is
                considered a neighbor.
            target (Element): Target element

        Returns:
            Sites coordinating input site.
        """
        coordinated_sites = []
        for site, weight in self.get_voronoi_polyhedra(n).items():
            if weight > tol and (target is None or site.specie == target):
                coordinated_sites.append(site)
        return coordinated_sites


class RelaxationAnalyzer(object):
    """
    This class analyzes the relaxation in a calculation.
    """

    def __init__(self, initial_structure, final_structure):
        """
        Please note that the input and final structures should have the same
        ordering of sites. This is typically the case for most computational
        codes.

        Args:
            initial_structure (Structure): Initial input structure to
                calculation.
            final_structure (Structure): Final output structure from
                calculation.
        """
        if final_structure.formula != initial_structure.formula:
            raise ValueError("Initial and final structures have different " +
                             "formulas!")
        self.initial = initial_structure
        self.final = final_structure

    def get_percentage_volume_change(self):
        """
        Returns the percentage volume change.

        Returns:
            Volume change in percentage, e.g., 0.055 implies a 5.5% increase.
        """
        initial_vol = self.initial.lattice.volume
        final_vol = self.final.lattice.volume
        return final_vol / initial_vol - 1

    def get_percentage_lattice_parameter_changes(self):
        """
        Returns the percentage lattice parameter changes.

        Returns:
            A dict of the percentage change in lattice parameter, e.g.,
            {'a': 0.012, 'b': 0.021, 'c': -0.031} implies a change of 1.2%,
            2.1% and -3.1% in the a, b and c lattice parameters respectively.
        """
        initial_latt = self.initial.lattice
        final_latt = self.final.lattice
        d = {l: getattr(final_latt, l) / getattr(initial_latt, l) - 1
             for l in ["a", "b", "c"]}
        return d

    def get_percentage_bond_dist_changes(self, max_radius=3.0):
        """
        Returns the percentage bond distance changes for each site up to a
        maximum radius for nearest neighbors.

        Args:
            max_radius (float): Maximum radius to search for nearest
               neighbors. This radius is applied to the initial structure,
               not the final structure.

        Returns:
            Bond distance changes as a dict of dicts. E.g.,
            {index1: {index2: 0.011, ...}}. For economy of representation, the
            index1 is always less than index2, i.e., since bonding between
            site1 and siten is the same as bonding between siten and site1,
            there is no reason to duplicate the information or computation.
        """
        data = collections.defaultdict(dict)
        for inds in itertools.combinations(list(range(len(self.initial))), 2):
            (i, j) = sorted(inds)
            initial_dist = self.initial[i].distance(self.initial[j])
            if initial_dist < max_radius:
                final_dist = self.final[i].distance(self.final[j])
                data[i][j] = final_dist / initial_dist - 1
        return data


class VoronoiConnectivity(object):
    """
    Computes the solid angles swept out by the shared face of the voronoi
    polyhedron between two sites.

    Args:
        structure (Structure): Input structure
        cutoff (float) Cutoff distance.
    """

    # Radius in Angstrom cutoff to look for coordinating atoms

    def __init__(self, structure, cutoff=10):
        self.cutoff = cutoff
        self.s = structure
        recp_len = np.array(self.s.lattice.reciprocal_lattice.abc)
        i = np.ceil(cutoff * recp_len / (2 * math.pi))
        offsets = np.mgrid[-i[0]:i[0] + 1, -i[1]:i[1] + 1, -i[2]:i[2] + 1].T
        self.offsets = np.reshape(offsets, (-1, 3))
        # shape = [image, axis]
        self.cart_offsets = self.s.lattice.get_cartesian_coords(self.offsets)

    @property
    def connectivity_array(self):
        """
        Provides connectivity array.

        Returns:
            connectivity: An array of shape [atomi, atomj, imagej]. atomi is
            the index of the atom in the input structure. Since the second
            atom can be outside of the unit cell, it must be described
            by both an atom index and an image index. Array data is the
            solid angle of polygon between atomi and imagej of atomj
        """
        # shape = [site, axis]
        cart_coords = np.array(self.s.cart_coords)
        # shape = [site, image, axis]
        all_sites = cart_coords[:, None, :] + self.cart_offsets[None, :, :]
        vt = Voronoi(all_sites.reshape((-1, 3)))
        n_images = all_sites.shape[1]
        cs = (len(self.s), len(self.s), len(self.cart_offsets))
        connectivity = np.zeros(cs)
        vts = np.array(vt.vertices)
        for (ki, kj), v in vt.ridge_dict.items():
            atomi = ki // n_images
            atomj = kj // n_images

            imagei = ki % n_images
            imagej = kj % n_images

            if imagei != n_images // 2 and imagej != n_images // 2:
                continue

            if imagei == n_images // 2:
                # atomi is in original cell
                val = solid_angle(vt.points[ki], vts[v])
                connectivity[atomi, atomj, imagej] = val

            if imagej == n_images // 2:
                # atomj is in original cell
                val = solid_angle(vt.points[kj], vts[v])
                connectivity[atomj, atomi, imagei] = val

            if -10.101 in vts[v]:
                warn('Found connectivity with infinite vertex. '
                     'Cutoff is too low, and results may be '
                     'incorrect')
        return connectivity

    @property
    def max_connectivity(self):
        """
        returns the 2d array [sitei, sitej] that represents
        the maximum connectivity of site i to any periodic
        image of site j
        """
        return np.max(self.connectivity_array, axis=2)

    def get_connections(self):
        """
        Returns a list of site pairs that are Voronoi Neighbors, along
        with their real-space distances.
        """
        con = []
        maxconn = self.max_connectivity
        for ii in range(0, maxconn.shape[0]):
            for jj in range(0, maxconn.shape[1]):
                if maxconn[ii][jj] != 0:
                    dist = self.s.get_distance(ii, jj)
                    con.append([ii, jj, dist])
        return con

    def get_sitej(self, site_index, image_index):
        """
        Assuming there is some value in the connectivity array at indices
        (1, 3, 12). sitei can be obtained directly from the input structure
        (structure[1]). sitej can be obtained by passing 3, 12 to this function

        Args:
            site_index (int): index of the site (3 in the example)
            image_index (int): index of the image (12 in the example)
        """
        atoms_n_occu = self.s[site_index].species_and_occu
        lattice = self.s.lattice
        coords = self.s[site_index].frac_coords + self.offsets[image_index]
        return PeriodicSite(atoms_n_occu, coords, lattice)


def solid_angle(center, coords):
    """
    Helper method to calculate the solid angle of a set of coords from the
    center.

    Args:
        center (3x1 array): Center to measure solid angle from.
        coords (Nx3 array): List of coords to determine solid angle.

    Returns:
        The solid angle.
    """
    o = np.array(center)
    r = [np.array(c) - o for c in coords]
    r.append(r[0])
    n = [np.cross(r[i + 1], r[i]) for i in range(len(r) - 1)]
    n.append(np.cross(r[1], r[0]))
    vals = []
    for i in range(len(n) - 1):
        v = -np.dot(n[i], n[i + 1]) \
            / (np.linalg.norm(n[i]) * np.linalg.norm(n[i + 1]))
        vals.append(math.acos(abs_cap(v)))
    phi = sum(vals)
    return phi + (3 - len(r)) * math.pi


def contains_peroxide(structure, relative_cutoff=1.1):
    """
    Determines if a structure contains peroxide anions.

    Args:
        structure (Structure): Input structure.
        relative_cutoff: The peroxide bond distance is 1.49 Angstrom.
            Relative_cutoff * 1.49 stipulates the maximum distance two O
            atoms must be to each other to be considered a peroxide.

    Returns:
        Boolean indicating if structure contains a peroxide anion.
    """
    ox_type = oxide_type(structure, relative_cutoff)
    if ox_type == "peroxide":
        return True
    else:
        return False


class OxideType(object):
    """
    Separate class for determining oxide type.

    Args:
        structure: Input structure.
        relative_cutoff: Relative_cutoff * act. cutoff stipulates the max.
            distance two O atoms must be from each other. Default value is
            1.1. At most 1.1 is recommended, nothing larger, otherwise the
            script cannot distinguish between superoxides and peroxides.
    """

    def __init__(self, structure, relative_cutoff=1.1):
        self.structure = structure
        self.relative_cutoff = relative_cutoff
        self.oxide_type, self.nbonds = self.parse_oxide()

    def parse_oxide(self):
        """
        Determines if an oxide is a peroxide/superoxide/ozonide/normal oxide.

        Returns:
            oxide_type (str): Type of oxide
            ozonide/peroxide/superoxide/hydroxide/None.
            nbonds (int): Number of peroxide/superoxide/hydroxide bonds in
            structure.
        """
        structure = self.structure
        relative_cutoff = self.relative_cutoff
        o_sites_frac_coords = []
        h_sites_frac_coords = []
        lattice = structure.lattice

        if isinstance(structure.composition.elements[0], Element):
            comp = structure.composition
        elif isinstance(structure.composition.elements[0], Specie):
            elmap = collections.defaultdict(float)
            for site in structure:
                for species, occu in site.species_and_occu.items():
                    elmap[species.element] += occu
            comp = Composition(elmap)
        if Element("O") not in comp or comp.is_element:
            return "None", 0

        for site in structure:
            syms = [sp.symbol for sp in site.species_and_occu.keys()]
            if "O" in syms:
                o_sites_frac_coords.append(site.frac_coords)
            if "H" in syms:
                h_sites_frac_coords.append(site.frac_coords)

        if h_sites_frac_coords:
            dist_matrix = lattice.get_all_distances(o_sites_frac_coords,
                                                    h_sites_frac_coords)
            if np.any(dist_matrix < relative_cutoff * 0.93):
                return "hydroxide", len(
                    np.where(dist_matrix < relative_cutoff * 0.93)[0]) / 2.0
        dist_matrix = lattice.get_all_distances(o_sites_frac_coords,
                                                o_sites_frac_coords)
        np.fill_diagonal(dist_matrix, 1000)
        is_superoxide = False
        is_peroxide = False
        is_ozonide = False
        if np.any(dist_matrix < relative_cutoff * 1.35):
            bond_atoms = np.where(dist_matrix < relative_cutoff * 1.35)[0]
            is_superoxide = True
        elif np.any(dist_matrix < relative_cutoff * 1.49):
            is_peroxide = True
            bond_atoms = np.where(dist_matrix < relative_cutoff * 1.49)[0]
        if is_superoxide:
            if len(bond_atoms) > len(set(bond_atoms)):
                is_superoxide = False
                is_ozonide = True
        try:
            nbonds = len(set(bond_atoms))
        except UnboundLocalError:
            nbonds = 0.0
        if is_ozonide:
            str_oxide = "ozonide"
        elif is_superoxide:
            str_oxide = "superoxide"
        elif is_peroxide:
            str_oxide = "peroxide"
        else:
            str_oxide = "oxide"
        if str_oxide == "oxide":
            nbonds = comp["O"]
        return str_oxide, nbonds


def oxide_type(structure, relative_cutoff=1.1, return_nbonds=False):
    """
    Determines if an oxide is a peroxide/superoxide/ozonide/normal oxide

    Args:
        structure (Structure): Input structure.
        relative_cutoff (float): Relative_cutoff * act. cutoff stipulates the
            max distance two O atoms must be from each other.
        return_nbonds (bool): Should number of bonds be requested?
    """

    ox_obj = OxideType(structure, relative_cutoff)
    if return_nbonds:
        return ox_obj.oxide_type, ox_obj.nbonds
    else:
        return ox_obj.oxide_type


def gramschmidt(vin, uin):
    """
    Returns that part of the first input vector
    that is orthogonal to the second input vector.
    The output vector is not normalized.

    Args:
        vin (numpy array):
            first input vector
        uin (numpy array):
            second input vector
    """

    vin_uin = np.inner(vin, uin)
    uin_uin = np.inner(uin, uin)
    if uin_uin <= 0.0:
        raise ValueError("Zero or negative inner product!")
    return vin - (vin_uin / uin_uin) * uin


class OrderParameters(object):
    """
    This class permits the calculation of various types of local order
    parameters.
    """

    __supported_types = ("cn", "tet", "oct", "bcc", "q2", "q4", "q6")

    def __init__(self, types, parameters=None, cutoff=-10.0):
        """
        Create an OrderParameter analyzer instance.

        Args:
            types ([string]):
                List of strings representing the types of order parameters
                to be calculated. Note that multiple mentions of the
                same type may occur. Currently available types are
                "cn"  (simple coordination number---normalized,
                      if desired),
                "tet" [Shetty--Peters style OP recognizing tetrahedral
                      coordination (Zimmermann et al.,
                      J. Am. Chem. Soc., 137, 13352-13361, 2015)],
                "oct" [Shetty--Peters style OP recognizing octahedral
                      coordination (Zimmermann et al.,
                      J. Am. Chem. Soc., 137, 13352-13361, 2015)],
                "bcc" [Shetty--Peters style OP recognizing local
                      body-centered cubic environment (Peters,
                      J. Chem. Phys., 131, 244103, 2009)],
                "q2"  [Bond orientational order parameter (BOOP)
                      of weight l=2 (Steinhardt et al., Phys. Rev. B,
                      28, 784-805, 1983)],
                "q4"  (BOOP of weight l=4),
                "q6"  (BOOP of weight l=6).
            parameters ([[float]]):
                2D list of floating point numbers that store
                parameters associated with the different order parameters
                that are to be calculated (1st dimension = length of
                types tuple; any 2nd dimension may be zero, in which case
                default values are used). In the following, those order
                parameters q_i are listed that require further parameters
                for their computation (values in brackets denote default
                values):
                  "cn":  normalizing constant (1);
                  "tet": Gaussian width in fractions of pi (180 degrees)
                         reflecting the "speed of
                         penalizing" deviations away from the perfect
                         tetrahedral angle of any individual
                         neighbor1-center-neighbor2 configuration (0.0667);
                  "oct": threshold angle in degrees distinguishing a second
                         neighbor to be either close to the south pole or
                         close to the equator (160.0);
                         Gaussian width for penalizing deviations away
                         from south pole (0.0667);
                         Gaussian width for penalizing deviations away
                         from equator (0.0556);
                         constant for shifting q_oct toward smaller
                         values, which can be helpful when trying to fine-
                         tune the capabilities of distinguishing between
                         different environments (e.g., tet vs oct)
                         given a single mutual threshold q_thresh;
                  "bcc": south-pole threshold angle as for "oct" (160.0);
                         south-pole Gaussian width as for "oct" (0.0667).
            cutoff (float):
                Cutoff radius to determine which nearest neighbors are
                supposed to contribute to the order parameters.
                If the value is negative the neighboring sites found by
                distance and cutoff radius are further
                pruned using the get_coordinated_sites method from the
                VoronoiCoordFinder class.
        """
        if len(types) == 0:
            raise ValueError("Empty types list!")
        for t in types:
            if t not in OrderParameters.__supported_types:
                raise ValueError("Unknown order parameter type (" + \
                                 t + ")!")
        if parameters:
            if len(types) != len(parameters):
                raise ValueError("1st dimension of parameters array is not"
                                 " consistent with types list!")
            for lp in parameters:
                if len(lp) > 0:
                    for p in lp:
                        if type(p) != float and type(p) != int:
                            raise AttributeError("Expected only float and"
                                                 " integer type parameters!")
            loc_parameters = list(parameters)
        else:
            loc_parameters = [[] for t in types]
        self._types = tuple(types)
        tmpparas = []
        self._computerijs = self._geomops = self._boops = False
        self._max_trig_order = -1
        for i, t in enumerate(self._types):
            # add here any additional parameter checking and
            #     default value assignment
            tmpparas.append([])
            if t == "cn":
                if len(loc_parameters[i]) == 0:
                    tmpparas[i].append(1.0)
                else:
                    if loc_parameters[i][0] == 0.0:
                        raise ValueError("Normalizing constant for"
                                         " coordination-number based order"
                                         " parameter is zero!")
                    else:
                        tmpparas[i].append(loc_parameters[i][0])
            elif t == "tet":
                if len(loc_parameters[i]) == 0:
                    tmpparas[i].append(1.0 / 0.0667)
                else:
                    if loc_parameters[i][0] == 0.0:
                        raise ValueError("Gaussian width for"
                                         " tetrahedral order"
                                         " parameter is zero!")
                    else:
                        tmpparas[i].append(1.0 / loc_parameters[i][0])
            elif t == "oct":
                if len(loc_parameters[i]) < 4:
                    tmpparas[i].append(8.0 * pi / 9.0)
                    tmpparas[i].append(1.0 / 0.0667)
                    tmpparas[i].append(1.0 / 0.0556)
                    tmpparas[i].append(0.25)
                    tmpparas[i].append(4.0 / 3.0)
                else:
                    if loc_parameters[i][0] <= 0.0 or loc_parameters[i][
                            0] >= 180.0:
                        warn("Threshold value for south pole"
                             " configurations in octahedral order"
                             " parameter outside ]0,180[")
                    tmpparas[i].append(loc_parameters[i][0] * pi / 180.0)
                    if loc_parameters[i][1] == 0.0:
                        raise ValueError("Gaussian width for south pole"
                                         " configurations in octahedral"
                                         " order parameter is zero!")
                    else:
                        tmpparas[i].append(1.0 / loc_parameters[i][1])
                    if loc_parameters[i][2] == 0.0:
                        raise ValueError("Gaussian width for equatorial"
                                         " configurations in octahedral"
                                         " order parameter is zero!")
                    else:
                        tmpparas[i].append(1.0 / loc_parameters[i][2])
                    if loc_parameters[i][3] - 1.0 == 0.0:
                        raise ValueError("Shift constant may not be"
                                         " unity!")
                    if loc_parameters[i][3] < 0.0 or loc_parameters[i][3] > 1.0:
                        warn("Shift constant outside [0,1[.")
                    tmpparas[i].append(loc_parameters[i][3])
                    tmpparas[i].append(1.0 / (1.0 - loc_parameters[i][3]))
            elif t == "bcc":
                if len(loc_parameters[i]) < 2:
                    tmpparas[i].append(8.0 * pi / 9.0)
                    tmpparas[i].append(1.0 / 0.0667)
                else:
                    if loc_parameters[i][0] <= 0.0 or loc_parameters[i][
                            0] >= 180.0:
                        warn("Threshold value for south pole"
                             " configurations in bcc order"
                             " parameter outside ]0,180[")
                    tmpparas[i].append(loc_parameters[i][0] * pi / 180.0)
                    if loc_parameters[i][1] == 0.0:
                        raise ValueError("Gaussian width for south pole"
                                         " configurations in bcc"
                                         " order parameter is zero!")
                    else:
                        tmpparas[i].append(1.0 / loc_parameters[i][1])
            # All following types should be well-defined/-implemented,
            # and they should not require parameters.
            elif t != "q2" and t != "q4" and t != "q6":
                raise ValueError("unknown order-parameter type \"" + t + "\"")

            # Add here any additional flags to be used during calculation.
            if t == "tet" or t == "oct" or t == "bcc":
                self._computerijs = self._geomops = True
            if t == "q2" or t == "q4" or t == "q6":
                self._computerijs = self._boops = True
            if t == "q2" and self._max_trig_order < 2:
                self._max_trig_order = 2
            if t == "q4" and self._max_trig_order < 4:
                self._max_trig_order = 4
            if t == "q6" and self._max_trig_order < 6:
                self._max_trig_order = 6

        # Finish parameter treatment.
        self._paras = list(tmpparas)
        if cutoff < 0.0:
            self._cutoff = -cutoff
            self._voroneigh = True
        elif cutoff > 0.0:
            self._cutoff = cutoff
            self._voroneigh = False
        else:
            raise ValueError("Cutoff radius is zero!")

        # Further variable definitions.
        self._last_nneigh = -1
        self._pow_sin_t = {}
        self._pow_cos_t = {}
        self._sin_n_p = {}
        self._cos_n_p = {}

    @property
    def num_ops(self):

        """"
        Returns the number of different order parameters that are targeted
        to be calculated.
        """

        return len(self._types)

    @property
    def last_nneigh(self):

        """"
        Returns the number of neighbors encountered during the most
        recent order-parameter calculation. A value of -1 indicates that
        no such calculation has yet been performed for this instance.
        """

        return len(self._last_nneigh)

    def compute_trigonometric_terms(self, thetas=[], phis=[]):

        """"
        Computes trigonometric terms that are required to
        calculate bond orientational order parameters.

        Args:
            thetas ([float]):
                polar angles of all neighbors in radians.
            phis ([float]):
                azimuth angles of all neighbors in radians.  The list of
                azimuth angles is expected to have the same size as the list
                of polar angles; otherwise, a ValueError is raised.  Also,
                the two lists of angles have to be coherent in order. That
                is, it is expected that the order in the list of azimuth
                angles corresponds to a distinct sequence of neighbors.
                And, this sequence has to equal the sequence
                of neighbors in the list of polar angles.

        """

        if not thetas or not phis:
            raise ValueError("No entries in list/s of polar and/or azimuthal" \
                             " angles!")
        if len(thetas) != len(phis):
            raise ValueError("List of polar and azimuthal angles have to be"
                             " equal!")

        self._pow_sin_t.clear()
        self._pow_cos_t.clear()
        self._sin_n_p.clear()
        self._cos_n_p.clear()

        self._pow_sin_t[1] = [math.sin(float(t)) for t in thetas]
        self._pow_cos_t[1] = [math.cos(float(t)) for t in thetas]
        self._sin_n_p[1] = [math.sin(float(p)) for p in phis]
        self._cos_n_p[1] = [math.cos(float(p)) for p in phis]

        for i in range(2, self._max_trig_order + 1):
            self._pow_sin_t[i] = [e[0] * e[1] for e in zip(
                self._pow_sin_t[i - 1], self._pow_sin_t[1])]
            self._pow_cos_t[i] = [e[0] * e[1] for e in zip(
                self._pow_cos_t[i - 1], self._pow_cos_t[1])]
            self._sin_n_p[i] = [math.sin(float(i) * float(p)) \
                                for p in phis]
            self._cos_n_p[i] = [math.cos(float(i) * float(p)) \
                                for p in phis]

    def get_q2(self, thetas=[], phis=[]):

        """
        Calculates the value of the bond orientational order parameter of
        weight l=2.  If the function is called with non-empty lists of
        polar and azimuthal angles the corresponding trigonometric terms
        are computed afresh.  Otherwise, it is expected that the
        compute_trigonometric_terms function has been just called.

        Args:
            thetas ([float]):
                polar angles of all neighbors in radians.
            phis ([float]):
                azimuth angles of all neighbors in radians.

        Return:
            q2 (float): bond orientational order parameter of weight l=2
                corresponding to the input angles thetas and phis.
        """

        if thetas and phis:
            self.compute_trigonometric_terms(thetas, phis)
        nnn = len(self._pow_sin_t[1])
        nnn_range = range(nnn)

        sqrt_15_2pi = math.sqrt(15.0 / (2.0 * pi))
        sqrt_5_pi = math.sqrt(5.0 / pi)

        pre_y_2_2 = [0.25 * sqrt_15_2pi * val for val in self._pow_sin_t[2]]
        pre_y_2_1 = [0.5 * sqrt_15_2pi * val[0] * val[1]
                     for val in zip(self._pow_sin_t[1], self._pow_cos_t[1])]

        acc = 0.0

        # Y_2_-2
        real = imag = 0.0
        for i in nnn_range:
            real += pre_y_2_2[i] * self._cos_n_p[2][i]
            imag -= pre_y_2_2[i] * self._sin_n_p[2][i]
        acc += (real * real + imag * imag)

        # Y_2_-1
        real = imag = 0.0
        for i in nnn_range:
            real += pre_y_2_1[i] * self._cos_n_p[1][i]
            imag -= pre_y_2_1[i] * self._sin_n_p[1][i]
        acc += (real * real + imag * imag)

        # Y_2_0
        real = imag = 0.0
        for i in nnn_range:
            real += 0.25 * sqrt_5_pi * (3.0 * self._pow_cos_t[2][i] - 1.0)
        acc += (real * real)

        # Y_2_1
        real = imag = 0.0
        for i in nnn_range:
            real -= pre_y_2_1[i] * self._cos_n_p[1][i]
            imag -= pre_y_2_1[i] * self._sin_n_p[1][i]
        acc += (real * real + imag * imag)

        # Y_2_2
        real = imag = 0.0
        for i in nnn_range:
            real += pre_y_2_2[i] * self._cos_n_p[2][i]
            imag += pre_y_2_2[i] * self._sin_n_p[2][i]
        acc += (real * real + imag * imag)

        q2 = math.sqrt(4.0 * pi * acc / (5.0 * float(nnn * nnn)))
        return q2

    def get_q4(self, thetas=[], phis=[]):

        """
        Calculates the value of the bond orientational order parameter of
        weight l=4.  If the function is called with non-empty lists of
        polar and azimuthal angles the corresponding trigonometric terms
        are computed afresh.  Otherwise, it is expected that the
        compute_trigonometric_terms function has been just called.

        Args:
            thetas ([float]):
                polar angles of all neighbors in radians.
            phis ([float]):
                azimuth angles of all neighbors in radians.

        Return:
            q4 (float): bond orientational order parameter of weight l=4
                corresponding to the input angles thetas and phis.
        """

        if thetas and phis:
            self.compute_trigonometric_terms(thetas, phis)
        nnn = len(self._pow_sin_t[1])
        nnn_range = range(nnn)

        i16_3 = 3.0 / 16.0
        i8_3 = 3.0 / 8.0

        sqrt_35_pi = math.sqrt(35.0 / pi)
        sqrt_35_2pi = math.sqrt(35.0 / (2.0 * pi))
        sqrt_5_pi = math.sqrt(5.0 / pi)
        sqrt_5_2pi = math.sqrt(5.0 / (2.0 * pi))
        sqrt_1_pi = math.sqrt(1.0 / pi)

        pre_y_4_4 = [i16_3 * sqrt_35_2pi * val for val in self._pow_sin_t[4]]
        pre_y_4_3 = [i8_3 * sqrt_35_pi * val[0] * val[1] \
                     for val in zip(self._pow_sin_t[3], self._pow_cos_t[1])]
        pre_y_4_2 = [i8_3 * sqrt_5_2pi * val[0] * (7.0 * val[1] - 1.0) \
                     for val in zip(self._pow_sin_t[2], self._pow_cos_t[2])]
        pre_y_4_1 = [i8_3 * sqrt_5_pi * val[0] * (7.0 * val[1] - 3.0 * val[2]) \
                     for val in zip(self._pow_sin_t[1], self._pow_cos_t[3], \
                                    self._pow_cos_t[1])]

        acc = 0.0

        # Y_4_-4
        real = imag = 0.0
        for i in nnn_range:
            real += pre_y_4_4[i] * self._cos_n_p[4][i]
            imag -= pre_y_4_4[i] * self._sin_n_p[4][i]
        acc += (real * real + imag * imag)

        # Y_4_-3
        real = imag = 0.0
        for i in nnn_range:
            real += pre_y_4_3[i] * self._cos_n_p[3][i]
            imag -= pre_y_4_3[i] * self._sin_n_p[3][i]
        acc += (real * real + imag * imag)

        # Y_4_-2
        real = imag = 0.0
        for i in nnn_range:
            real += pre_y_4_2[i] * self._cos_n_p[2][i]
            imag -= pre_y_4_2[i] * self._sin_n_p[2][i]
        acc += (real * real + imag * imag)

        # Y_4_-1
        real = imag = 0.0
        for i in nnn_range:
            real += pre_y_4_1[i] * self._cos_n_p[1][i]
            imag -= pre_y_4_1[i] * self._sin_n_p[1][i]
        acc += (real * real + imag * imag)

        # Y_4_0
        real = imag = 0.0
        for i in nnn_range:
            real += i16_3 * sqrt_1_pi * (35.0 * self._pow_cos_t[4][i] - \
                                         30.0 * self._pow_cos_t[2][i] + 3.0)
        acc += (real * real)

        # Y_4_1
        real = imag = 0.0
        for i in nnn_range:
            real -= pre_y_4_1[i] * self._cos_n_p[1][i]
            imag -= pre_y_4_1[i] * self._sin_n_p[1][i]
        acc += (real * real + imag * imag)

        # Y_4_2
        real = imag = 0.0
        for i in nnn_range:
            real += pre_y_4_2[i] * self._cos_n_p[2][i]
            imag += pre_y_4_2[i] * self._sin_n_p[2][i]
        acc += (real * real + imag * imag)

        # Y_4_3
        real = imag = 0.0
        for i in nnn_range:
            real -= pre_y_4_3[i] * self._cos_n_p[3][i]
            imag -= pre_y_4_3[i] * self._sin_n_p[3][i]
        acc += (real * real + imag * imag)

        # Y_4_4
        real = imag = 0.0
        for i in nnn_range:
            real += pre_y_4_4[i] * self._cos_n_p[4][i]
            imag += pre_y_4_4[i] * self._sin_n_p[4][i]
        acc += (real * real + imag * imag)

        q4 = math.sqrt(4.0 * pi * acc / (9.0 * float(nnn * nnn)))
        return q4

    def get_q6(self, thetas=[], phis=[]):

        """
        Calculates the value of the bond orientational order parameter of
        weight l=6.  If the function is called with non-empty lists of
        polar and azimuthal angles the corresponding trigonometric terms
        are computed afresh.  Otherwise, it is expected that the
        compute_trigonometric_terms function has been just called.

        Args:
            thetas ([float]):
                polar angles of all neighbors in radians.
            phis ([float]):
                azimuth angles of all neighbors in radians.

        Return:
            q6 (float): bond orientational order parameter of weight l=6
                corresponding to the input angles thetas and phis.
        """

        if thetas and phis:
            self.compute_trigonometric_terms(thetas, phis)
        nnn = len(self._pow_sin_t[1])
        nnn_range = range(nnn)

        i64 = 1.0 / 64.0
        i32 = 1.0 / 32.0
        i32_3 = 3.0 / 32.0
        i16 = 1.0 / 16.0

        sqrt_3003_pi = math.sqrt(3003.0 / pi)
        sqrt_1001_pi = math.sqrt(1001.0 / pi)
        sqrt_91_2pi = math.sqrt(91.0 / (2.0 * pi))
        sqrt_1365_pi = math.sqrt(1365.0 / pi)
        sqrt_273_2pi = math.sqrt(273.0 / (2.0 * pi))
        sqrt_13_pi = math.sqrt(13.0 / pi)

        pre_y_6_6 = [i64 * sqrt_3003_pi * val for val in self._pow_sin_t[6]]
        pre_y_6_5 = [i32_3 * sqrt_1001_pi * val[0] * val[1] \
                     for val in zip(self._pow_sin_t[5], self._pow_cos_t[1])]
        pre_y_6_4 = [i32_3 * sqrt_91_2pi * val[0] * (11.0 * val[1] - 1.0) \
                     for val in zip(self._pow_sin_t[4], self._pow_cos_t[2])]
        pre_y_6_3 = [
            i32 * sqrt_1365_pi * val[0] * (11.0 * val[1] - 3.0 * val[2]) \
            for val in zip(self._pow_sin_t[3], self._pow_cos_t[3], \
                           self._pow_cos_t[1])]
        pre_y_6_2 = [i64 * sqrt_1365_pi * val[0] * (33.0 * val[1] - \
                                                    18.0 * val[2] + 1.0) for val
                     in zip(self._pow_sin_t[2], \
                            self._pow_cos_t[4], self._pow_cos_t[2])]
        pre_y_6_1 = [i16 * sqrt_273_2pi * val[0] * (33.0 * val[1] - \
                                                    30.0 * val[2] + 5.0 * val[
                                                        3]) for val in
                     zip(self._pow_sin_t[1], \
                         self._pow_cos_t[5], self._pow_cos_t[3],
                         self._pow_cos_t[1])]

        acc = 0.0

        # Y_6_-6
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real += pre_y_6_6[i] * self._cos_n_p[6][i]  # cos(x) =  cos(-x)
            imag -= pre_y_6_6[i] * self._sin_n_p[6][i]  # sin(x) = -sin(-x)
        acc += (real * real + imag * imag)

        # Y_6_-5
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real += pre_y_6_5[i] * self._cos_n_p[5][i]
            imag -= pre_y_6_5[i] * self._sin_n_p[5][i]
        acc += (real * real + imag * imag)

        # Y_6_-4
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real += pre_y_6_4[i] * self._cos_n_p[4][i]
            imag -= pre_y_6_4[i] * self._sin_n_p[4][i]
        acc += (real * real + imag * imag)

        # Y_6_-3
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real += pre_y_6_3[i] * self._cos_n_p[3][i]
            imag -= pre_y_6_3[i] * self._sin_n_p[3][i]
        acc += (real * real + imag * imag)

        # Y_6_-2
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real += pre_y_6_2[i] * self._cos_n_p[2][i]
            imag -= pre_y_6_2[i] * self._sin_n_p[2][i]
        acc += (real * real + imag * imag)

        # Y_6_-1
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real += pre_y_6_1[i] * self._cos_n_p[1][i]
            imag -= pre_y_6_1[i] * self._sin_n_p[1][i]
        acc += (real * real + imag * imag)

        # Y_6_0
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real += i32 * sqrt_13_pi * (231.0 * self._pow_cos_t[6][i] - \
                                        315.0 * self._pow_cos_t[4][i] + 105.0 *
                                        self._pow_cos_t[2][i] - 5.0)
        acc += (real * real)

        # Y_6_1
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real -= pre_y_6_1[i] * self._cos_n_p[1][i]
            imag -= pre_y_6_1[i] * self._sin_n_p[1][i]
        acc += (real * real + imag * imag)

        # Y_6_2
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real += pre_y_6_2[i] * self._cos_n_p[2][i]
            imag += pre_y_6_2[i] * self._sin_n_p[2][i]
        acc += (real * real + imag * imag)

        # Y_6_3
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real -= pre_y_6_3[i] * self._cos_n_p[3][i]
            imag -= pre_y_6_3[i] * self._sin_n_p[3][i]
        acc += (real * real + imag * imag)

        # Y_6_4
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real += pre_y_6_4[i] * self._cos_n_p[4][i]
            imag += pre_y_6_4[i] * self._sin_n_p[4][i]
        acc += (real * real + imag * imag)

        # Y_6_5
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real -= pre_y_6_5[i] * self._cos_n_p[5][i]
            imag -= pre_y_6_5[i] * self._sin_n_p[5][i]
        acc += (real * real + imag * imag)

        # Y_6_6
        real = imag = 0.0
        real = 0.0
        imag = 0.0
        for i in nnn_range:
            real += pre_y_6_6[i] * self._cos_n_p[6][i]
            imag += pre_y_6_6[i] * self._sin_n_p[6][i]
        acc += (real * real + imag * imag)

        q6 = math.sqrt(4.0 * pi * acc / (13.0 * float(nnn * nnn)))
        return q6

    def get_type(self, index):

        """
        Return type of order-parameter at the index provided and
        represented by a short string.

        Args:
            index (int):
                index of order-parameter for which type is to be returned
        """
        if index < 0 or index >= len(self._types):
            raise ValueError("Index for getting order-parameter type"
                             " out-of-bounds!")
        return self._types[index]

    def get_parameters(self, index):

        """
        Returns list of floats that represents
        the parameters associated with calculation of the order
        parameter that was defined at the index provided.
        Attention: the parameters do not need to equal those originally
        inputted because of processing out of efficiency reasons.

        Args:
            index (int):
                index of order-parameter for which associated parameters
                are to be returned
        """
        if index < 0 or index >= len(self._types):
            raise ValueError("Index for getting parameters associated with"
                             " order-parameter calculation out-of-bounds!")
        return self._paras[index]

    def get_order_parameters(self, structure, n, indeces_neighs=[], \
                             tol=0.0, target_spec=None):

        """
        Compute all order parameters of site n.

        Args:
            structure (Structure):
                input structure.
            n (int):
                index of site in input structure, for which OPs are to be
                calculated.  Note that we do not use the sites iterator
                here, but directly access sites via struct[index].
            indeces_neighs ([int]):
                list of indeces of those neighbors in Structure object
                structure that are to be considered for OP computation.
                This optional argument overwrites the way neighbors are
                to be determined as defined in the constructor (i.e.,
                Voronoi coordination finder via negative cutoff radius
                vs constant cutoff radius if cutoff was positive).
                We do not use information about the underlying
                structure lattice if the neighbor indeces are explicitly
                provided.  This has two important consequences.  First,
                the input Structure object can, in fact, be a
                simple list of Site objects.  Second, no nearest images
                of neighbors are determined when providing an index list.
                Note furthermore that this neighbor
                determination type ignores the optional target_spec
                argument.
            tol (float):
                threshold of weight (= solid angle / maximal solid angle)
                to determine if a particular pair is
                considered neighbors; this is relevant only in the case
                when Voronoi polyhedra are used to determine coordination
            target_spec (Specie):
                target specie to be considered when calculating the order
                parameters of site n; None includes all species of input
                structure.

        Returns:
            list of floats representing order parameters.  Should it not be
            possible to compute a given OP for a conceptual reason, the
            corresponding entry is None instead of a float.  For Steinhardt
            et al.'s bond orientational OPs and the other geometric OPs
            ("tet", "oct", "bcc"), this can happen if there is a single
            neighbor around site n in the structure because that, obviously,
            does not permit calculation of angles between multiple
            neighbors.
        """

        # Do error-checking and initialization.
        if n < 0:
            raise ValueError("Site index smaller zero!")
        if n >= len(structure):
            raise ValueError("Site index beyond maximum!")
        if indeces_neighs:
            for index in indeces_neighs:
                if index >= len(structure):
                    raise ValueError("Neighbor site index beyond maximum!")
        if tol < 0.0:
            raise ValueError("Negative tolerance for weighted solid angle!")
        left_of_unity = 1.0 - 1.0e-12

        # Find central site and its neighbors.
        # Note that we adopt the same way of accessing sites here as in
        # VoronoiCoordFinder; that is, not via the sites iterator.
        centsite = structure[n]
        if indeces_neighs:
            neighsites = [structure[index] for index in indeces_neighs]
        elif self._voroneigh:
            vorocf = VoronoiCoordFinder(structure)
            neighsites = vorocf.get_coordinated_sites(n, tol, target_spec)
        else:
            # Structure.get_sites_in_sphere --> also other periodic images
            neighsitestmp = [i[0] for i in structure.get_sites_in_sphere(
                centsite.coords, self._cutoff)]
            neighsites = []
            if centsite not in neighsitestmp:
                raise ValueError("Could not find center site!")
            else:
                neighsitestmp.remove(centsite)
            if target_spec is None:
                neighsites = list(neighsitestmp)
            else:
                neighsites[:] = [site for site in neighsitestmp \
                                 if site.specie.symbol == target_spec]
        nneigh = len(neighsites)
        self._last_nneigh = nneigh

        # Prepare angle calculations, if applicable.
        rij = []
        rijnorm = []
        dist = []
        centvec = centsite.coords
        if self._computerijs:
            for j, neigh in enumerate(neighsites):
                rij.append((neigh.coords - centvec))
                dist.append(np.linalg.norm(rij[j]))
                rijnorm.append((rij[j] / dist[j]))

        # Initialize OP list and, then, calculate OPs.
        ops = [0.0 for t in self._types]

        # First, coordination number-based OPs.
        for i, t in enumerate(self._types):
            if t == "cn":
                ops[i] = nneigh / self._paras[i][0]

        # Then, bond orientational OPs based on spherical harmonics
        # according to Steinhardt et al., Phys. Rev. B, 28, 784-805, 1983.
        if self._boops:
            thetas = []
            phis = []
            for j, vec in enumerate(rijnorm):

                # z is North pole --> theta between vec and (0, 0, 1)^T.
                # Because vec is normalized, dot product is simply vec[2].
                thetas.append(math.acos(max(-1.0, min(vec[2], 1.0))))
                tmpphi = 0.0

                # Compute phi only if it is not (almost) perfectly
                # aligned with z-axis.
                if vec[2] < left_of_unity and vec[2] > - (left_of_unity):
                    # x is prime meridian --> phi between projection of vec
                    # into x-y plane and (1, 0, 0)^T
                    tmpphi = math.acos(max(
                        -1.0,
                        min(vec[0] / (math.sqrt(
                            vec[0] * vec[0] + vec[1] * vec[1])),
                            1.0)))
                    if vec[1] < 0.0:
                        tmpphi = -tmpphi
                phis.append(tmpphi)

            # Note that None flags that we have too few neighbors
            # for calculating BOOPS.
            for i, t in enumerate(self._types):
                if t == "q2":
                    ops[i] = self.get_q2(thetas, phis) if len(
                        thetas) > 0 else None
                elif t == "q4":
                    ops[i] = self.get_q4(thetas, phis) if len(
                        thetas) > 0 else None
                elif t == "q6":
                    ops[i] = self.get_q6(thetas, phis) if len(
                        thetas) > 0 else None

        # Then, deal with the Shetty--Peters style OPs that are tailor-made
        # to recognize common structural motifs
        # (Shetty et al., J. Chem. Phys., 117, 4000-4009, 2002;
        #  Peters, J. Chem. Phys., 131, 244103, 2009;
        #  Zimmermann et al., J. Am. Chem. Soc., under revision, 2015).
        if self._geomops:
            gaussthetak = [0.0 for t in self._types]  # not used by all OPs
            ipi = 1.0 / pi
            piover2 = pi / 2.0
            tetangoverpi = math.acos(-1.0 / 3.0) * ipi
            itetangminuspihalfoverpi = 1.0 / (tetangoverpi - 0.5)

            for j in range(nneigh):  # Neighbor j is put to the North pole.
                zaxis = rijnorm[j]

                for k in range(nneigh):  # From neighbor k, we construct
                    if j != k:  # the prime meridian.
                        tmp = max(
                            -1.0, min(np.inner(zaxis, rijnorm[k]), 1.0))
                        thetak = math.acos(tmp)
                        xaxistmp = gramschmidt(rijnorm[k], zaxis)
                        if np.linalg.norm(xaxistmp) == 0.0:
                            flag_xaxis = True
                        else:
                            xaxis = xaxistmp / np.linalg.norm(xaxistmp)
                            flag_xaxis = False

                        # Contributions of j-i-k angles, where i represents the central atom
                        # and j and k two of the neighbors.
                        for i, t in enumerate(self._types):
                            if t == "tet":
                                tmp = self._paras[i][0] * (
                                    thetak * ipi - tetangoverpi)
                                gaussthetak[i] = math.exp(-0.5 * tmp * tmp)
                            elif t == "oct":
                                if thetak >= self._paras[i][0]:
                                    # k is south pole to j
                                    tmp = self._paras[i][1] * (
                                        thetak * ipi - 1.0)
                                    ops[i] += 3.0 * math.exp(-0.5 * tmp * tmp)
                            elif t == "bcc" and j < k:
                                if thetak >= self._paras[i][0]:
                                    # k is south pole to j
                                    tmp = self._paras[i][1] * (
                                        thetak * ipi - 1.0)
                                    ops[i] += 6.0 * math.exp(-0.5 * tmp * tmp)

                        for m in range(nneigh):
                            if (m != j) and (m != k):
                                tmp = max(
                                    -1.0, min(np.inner(zaxis, rijnorm[m]), 1.0))
                                thetam = math.acos(tmp)
                                xtwoaxistmp = gramschmidt(rijnorm[m], zaxis)
                                l = np.linalg.norm(xtwoaxistmp)
                                if l == 0:
                                    flag_xtwoaxis = True
                                else:
                                    xtwoaxis = xtwoaxistmp / l
                                    phi = math.acos(max(
                                        -1.0, min(np.inner(xtwoaxis, xaxis), 1.0)))
                                    flag_xtwoaxis = False

                                # Contributions of j-i-m angle and
                                # angles between plane j-i-k and i-m vector.
                                if not flag_xaxis and not flag_xtwoaxis:
                                    for i, t in enumerate(self._types):
                                        if t == "tet":
                                            tmp = self._paras[i][0] * (
                                                thetam * ipi - tetangoverpi)
                                            ops[i] += gaussthetak[i] * math.exp(
                                                -0.5 * tmp * tmp) * math.cos(
                                                3.0 * phi)
                                        elif t == "oct":
                                            if thetak < self._paras[i][0] and \
                                                            thetam < \
                                                            self._paras[i][0]:
                                                tmp = math.cos(2.0 * phi)
                                                tmp2 = self._paras[i][2] * (
                                                    thetam * ipi - 0.5)
                                                ops[i] += tmp * tmp * \
                                                          self._paras[i][4] * (
                                                              math.exp(
                                                                  -0.5 * tmp2 * tmp2) - \
                                                              self._paras[i][3])
                                        elif t == "bcc" and j < k:
                                            if thetak < self._paras[i][0]:
                                                if thetak > piover2:
                                                    fac = 1.0
                                                else:
                                                    fac = -1.0
                                                tmp = (thetam - piover2) / (
                                                19.47 * pi / 180.0)
                                                ops[i] += fac * math.cos(
                                                    3.0 * phi) * \
                                                          1.6 * tmp * \
                                                          math.exp(
                                                              -0.5 * tmp * tmp)

            # Normalize Shetty--Peters style OPs.
            for i, t in enumerate(self._types):
                if t == "tet":
                    ops[i] = ops[i] / 24.0 if nneigh > 2 else None

                elif t == "oct":
                    ops[i] = ops[i] / 90.0 if nneigh > 2 else None

                elif t == "bcc":
                    # Reassured 144 by pen-and-paper for perfect bcc
                    # structure (24 from i-j-k South pole contributions
                    # and 4 * 6 * 5 = 120 from i-j-k-m non-South
                    # pole-containing quadruples.
                    # Because qbcc has a separate South pole contribution
                    # as summand, it is sufficient to have 2 neighbors to
                    # obtain a configuration that is well-defined by
                    # yielding a contribution to qbcc via thetak alone:
                    # ==> nneigh > 1.
                    ops[i] = ops[i] / 144.0 if nneigh > 1 else None

        return ops



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals

from collections import namedtuple
from monty.json import MSONable

"""
A module to perform NMR data analysis/processing.
"""


__author__ = "Xiaohui Qu"
__copyright__ = "Copyright 2016, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Xiaohui Qu"
__email__ = "xhqu1981@gmail.com"
__date__ = "Apr 17, 2016"


class NMRChemicalShiftNotation(MSONable):
    """
    Helper class to convert between different chemical shift conventions
    internally using the Mehring notation. Note that this is different than the
    default notion adopted by VASP which is the Maryland notation.

    Three notations to describe chemical shift tensor (RK Harris; Magn. Reson.
    Chem. 2008, 46, 582–598; DOI: 10.1002/mrc.2225) are supported.

    Args:
        sigma_1 (float): chemical shift tensor principle component 1
        sigma_2 (float): chemical shift tensor principle component 2
        sigma_3 (float): chemical shift tensor principle component 3

    .. attribute:: sigma_11, simga_22, sigma33
        principle components in Mehring notation

    Authors: Xiaohui Qu
    """

    HaeberlenNotation = namedtuple(typename="HaeberlenNotion",
                                   field_names="sigma_iso, delta_sigma, zeta, eta")
    MehringNotation = namedtuple(typename="MehringNotation",
                                 field_names="sigma_iso, sigma_11, sigma_22, sigma_33")
    MarylandNotation = namedtuple(typename="MarylandNotation",
                                  field_names="sigma_iso, omega, kappa")

    def __init__(self, sigma_1, sigma_2, sigma_3):
        sigmas = sorted([sigma_1, sigma_2, sigma_3])
        self.sigma_11, self.sigma_22, self.sigma_33 = sigmas

    @property
    def haeberlen_values(self):
        """
        Returns: the Chemical shift tensor in Haeberlen Notation
        """
        sigma_iso = (self.sigma_11 + self.sigma_22 + self.sigma_33) / 3.0
        h_order_sigmas = sorted([self.sigma_11, self.sigma_22, self.sigma_33],
                                key=lambda x: abs(x - sigma_iso),
                                reverse=True)
        sigma_zz, sigma_xx, sigma_yy = h_order_sigmas
        delta_sigma = sigma_zz - 0.5 * (sigma_xx + sigma_yy)
        zeta = sigma_zz - sigma_iso
        assert abs(delta_sigma - 1.5 * zeta) < 1.0E-5
        eta = (sigma_yy - sigma_xx) / zeta
        return self.HaeberlenNotation(sigma_iso, delta_sigma, zeta, eta)

    @property
    def mehring_values(self):
        """
        Returns: the Chemical shift tensor in Mehring Notation
        """
        sigma_iso = (self.sigma_11 + self.sigma_22 + self.sigma_33) / 3.0
        return self.MehringNotation(sigma_iso, self.sigma_11,
                                    self.sigma_22, self.sigma_33)

    @property
    def maryland_values(self):
        """
        Returns: the Chemical shift tensor in Maryland Notation
        """
        sigma_iso = (self.sigma_11 + self.sigma_22 + self.sigma_33) / 3.0
        omega = self.sigma_33 - self.sigma_11
        # There is a typo in equation 20 from Magn. Reson. Chem. 2008, 46, 582–598, the sign is wrong.
        # There correct order is presented in Solid State Nucl. Magn. Reson. 1993, 2, 285-288.
        kappa = 3.0 * (self.sigma_22 - sigma_iso) / omega
        return self.MarylandNotation(sigma_iso, omega, kappa)

    @classmethod
    def from_maryland_notation(cls, sigma_iso, omega, kappa):
        sigma_22 = sigma_iso + kappa * omega / 3.0
        sigma_11 = (3.0 * sigma_iso - omega - sigma_22) / 2.0
        sigma_33 = 3.0 * sigma_iso - sigma_22 - sigma_11
        return cls(sigma_11, sigma_22, sigma_33)

    def as_dict(self):
        d = {"sigma_11": self.sigma_11,
             "sigma_22": self.sigma_22,
             "sigma_33": self.sigma_33}
        return d

    @classmethod
    def from_dict(cls, d):
        return cls(d["sigma_11"], d["sigma_22"], d["sigma_33"])



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, print_function, unicode_literals, \
    absolute_import

"""
This module defines classes that set the force field parameters for the bonds,
angles and dihedrals.

TODO: complete the class by defining interfaces to AMBER and CHARMM forcefields
"""

from collections import OrderedDict

from monty.json import MSONable

__author__ = 'Kiran Mathew, Navnidhi Rajput'


class ForceField(MSONable):
    """
    Stores force field information.

    Args:
        atoms (OrderedDict): store atomic mass for each atom name.
            { "atom name": atom mass, ... }
        bonds (OrderedDict): store the bond distance (A) and spring constant (
            Kcal/molA2) for each bond.
            { ("atom name1", "atom name2"): [spring const, distance], ... }
        angles (OrderedDict): store the bond angle and spring constant
            (Kcal/mol*radian2).
            { ("atom name1", "atom name2", "atom name3"): [spring const, angle], ... }
        dihedrals (OrderedDict): store the magnitude of torsion (Kcal/mol).
            { ("atom name1", "atom name2", "atom name3", "atom name4"): [
            function type, value, angle], ... }
        imdihedrals (OrderedDict): store improper dihedral information.
            similar to dihedrals but the gaff atom name1 and gaff atom name2 are marked 'X'
        vdws (OrderedDict): store the van der waal radius (A) and van der wall
            depth for a given atom (Kcal/mol). Lennard-Jones parameters.
            { "atom name": [sigma, epsilon], ... }
    """

    def __init__(self, atoms, bonds, angles, dihedrals=None, imdihedrals=None,
                 vdws=None, masses=None, charges=None):
        self.atoms = atoms
        self.bonds = bonds
        self.angles = angles
        self.dihedrals = dihedrals
        self.imdihedrals = imdihedrals
        self.vdws = vdws
        self.masses = masses
        self.charges = OrderedDict() if charges is None else charges
        self.atomindex_to_atomname = OrderedDict()
        self.atomindex_to_gaffname = OrderedDict()
        self.atomname_to_gaffname = OrderedDict()
        self.gaffname_to_atomindex = OrderedDict()
        # self.set_atom_mappings()

    # TODO: add interfaces to AMBER and CHARMM force fields


# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals, print_function

"""
Input sets for VASP GW calculations
Single vasp GW work: Creates input and jobscripts from the input sets for a specific job
"""


__author__ = "Michiel van Setten"
__copyright__ = " "
__version__ = "0.9"
__maintainer__ = "Michiel van Setten"
__email__ = "mjvansetten@gmail.com"
__date__ = "May 2014"

import os
import json
import os.path
import stat

from monty.serialization import loadfn

from pymatgen.io.vasp.inputs import Kpoints, Potcar
from pymatgen.io.vasp.sets import DictVaspInputSet
from pymatgen.io.abinit.helpers import s_name

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))

GWVaspInputSet = os.path.join(MODULE_DIR, "GWVaspInputSet.yaml")


"""
MPGWVaspInputSet.joson contains the standards for GW calculations. This set contains all
parameters for the first sc dft calculation. The modifications for the subsequent
sub calculations are made below.
For many settings the number of cores on which the calculations will be run is needed, this
number is assumed to be on the environment variable NPARGWCALC.
"""


class GWscDFTPrepVaspInputSet(DictVaspInputSet):
    """
    Implementation of VaspInputSet overriding MaterialsProjectVaspInputSet
    for static calculations preparing for a GW calculation.
    """
    TESTS = {}
    CONVS = {}

    def __init__(self, structure, spec, functional='PBE', sym_prec=0.01,
                 **kwargs):
        """
        Supports the same kwargs as :class:`JSONVaspInputSet`.
        """
        super(GWscDFTPrepVaspInputSet, self).__init__(
            "MP Static Self consistent run for GW",
            loadfn(GWVaspInputSet), **kwargs)
        self.structure = structure
        self.tests = self.__class__.get_defaults_tests()
        self.convs = self.__class__.get_defaults_convs()
        self.functional = functional
        self.set_dens(spec)
        self.sym_prec = sym_prec
    #  todo update the fromdict and todict ot include the new atributes

    @classmethod
    def get_defaults_tests(cls):
        return cls.TESTS.copy()

    @classmethod
    def get_defaults_convs(cls):
        return cls.CONVS.copy()

    def get_npar(self, structure):
        """
        get 'optimally' useful number of parallelism
        """
        npar = int(self.get_bands(structure) ** 2 * structure.volume / 200)
        npar = min(max(npar, 1), 52)
        return npar

    def set_test(self, test, value):
        """
        Method to switch a specific test on
        """
        all_tests = GWscDFTPrepVaspInputSet.get_defaults_tests()
        all_tests.update(GWDFTDiagVaspInputSet.get_defaults_tests())
        all_tests.update(GWG0W0VaspInputSet.get_defaults_tests())
        test_type = all_tests[test]['method']
        npar = self.get_npar(self.structure)
        if test_type == 'incar_settings':
            self.incar_settings.update({test: value})
        if test_type == 'set_nomega':
            nomega = npar * int(value / npar)
            self.incar_settings.update({"NOMEGA": int(nomega)})
        if test_type == 'set_nbands':
            nbands = value * self.get_bands(self.structure)
            nbands = npar * int(nbands / npar + 1)
            self.incar_settings.update({"NBANDS": int(nbands)})
        if test_type == 'kpoint_grid':
            pass

    def get_potcar(self, structure):
        """
        Method for getting LDA potcars
        """
        if self.sort_structure:
            structure = structure.get_sorted_structure()
        return Potcar(self.get_potcar_symbols(structure), functional=self.functional)

    def get_kpoints(self, structure):
        """
        Writes out a KPOINTS file using the automated gamma grid method.
        VASP crashes GW calculations on none gamma centered meshes.
        """
        if self.sort_structure:
            structure = structure.get_sorted_structure()
        dens = int(self.kpoints_settings['grid_density'])
        if dens == 1:
            return Kpoints.gamma_automatic()
        else:
            return Kpoints.automatic_gamma_density(structure, dens)

    def set_dens(self, spec):
        """
        sets the grid_density to the value specified in spec
        """
        self.kpoints_settings['grid_density'] = spec['kp_grid_dens']
        if spec['kp_grid_dens'] < 100:
            self.incar_settings.update({'ISMEAR': 0})

    def get_electrons(self, structure):
        """
        Method for retrieving the number of valence electrons
        """
        valence_list = {}
        potcar = self.get_potcar(structure)
        for pot_single in potcar:
            valence_list.update({pot_single.element: pot_single.nelectrons})
        electrons = sum([valence_list[element.symbol] for element in structure.species])
        return int(electrons)

    def get_bands(self, structure):
        """
        Method for retrieving the standard number of bands
        """
        bands = self.get_electrons(structure) / 2 + len(structure)
        return int(bands)

    def set_test_calc(self):
        """
        absolute minimal setting for testing
        """
        self.incar_settings.update({"PREC": "low", "ENCUT": 250})
        self.kpoints_settings['grid_density'] = 1

    def set_prec_high(self):
        self.incar_settings.update({"PREC": "Accurate", "ENCUT": 400})


class GWDFTDiagVaspInputSet(GWscDFTPrepVaspInputSet):
    """
    Implementation of VaspInputSet overriding MaterialsProjectVaspInputSet
    for static non self-consistent exact diagonalization step preparing for
    a GW calculation.
    """
    TESTS = {'NBANDS': {'test_range': (10, 20, 30), 'method': 'set_nbands', 'control': "gap"}}
    CONVS = {'NBANDS': {'test_range': (10, 20, 30, 40, 50, 60, 70), 'method': 'set_nbands', 'control': "gap"}}

    def __init__(self, structure, spec, functional='PBE', sym_prec=0.01, **kwargs):
        """
        Supports the same kwargs as :class:`JSONVaspInputSet`.
        """
        super(GWDFTDiagVaspInputSet, self).__init__(
            structure, spec, functional=functional, sym_prec=sym_prec, **kwargs)
        self.structure = structure
        self.tests = self.__class__.get_defaults_tests()
        self.convs = self.__class__.get_defaults_convs()
        self.functional = functional
        self.sym_prec = sym_prec
        self.set_dens(spec)
        npar = self.get_npar(self.structure)
        #single step exact diagonalization, output WAVEDER
        self.incar_settings.update({"ALGO": "Exact", "NELM": 1, "LOPTICS": "TRUE"})
        # for large systems exact diagonalization consumes too much memory
        self.set_gw_bands(15)
        self.incar_settings.update({"NPAR": npar})

    def set_gw_bands(self, factor=15):
        """
        method to set the number of bands for GW
        """
        gw_bands = self.get_bands(self.structure)
        gw_bands = self.get_npar(self.structure) * int((factor * gw_bands) / self.get_npar(self.structure) + 1)
        self.incar_settings.update({"NBANDS": gw_bands})
        if gw_bands > 800:
            self.incar_settings.update({"ALGO": 'fast'})

    def set_prec_high(self):
        super(GWDFTDiagVaspInputSet, self).set_prec_high()
        self.set_gw_bands(30)


class GWG0W0VaspInputSet(GWDFTDiagVaspInputSet):
    """
    Should go to Pymatgen vaspinputsets
    Implementation of VaspInputSet overriding MaterialsProjectVaspInputSet
    for static G0W0 calculation
    """
    TESTS = {'ENCUTGW': {'test_range': (200, 300, 400),
                         'method': 'incar_settings', 'control': "gap"},
             'NOMEGA': {'test_range': (80, 100, 120),
                        'method': 'set_nomega', 'control': "gap"}}
    CONVS = {'ENCUTGW': {'test_range': (200, 400, 600, 800),
                         'method': 'incar_settings', 'control': "gap"}}

    def __init__(self, structure, spec, functional='PBE', sym_prec=0.01,
                 **kwargs):
        """
        Supports the same kwargs as :class:`JSONVaspInputSet`.
        """
        super(GWG0W0VaspInputSet, self).__init__(
            structure, spec, functional=functional, sym_prec=sym_prec, **kwargs)
        self.structure = structure
        self.tests = self.__class__.get_defaults_tests()
        self.convs = self.__class__.get_defaults_convs()
        self.functional = functional
        self.sym_prec = sym_prec
        npar = self.get_npar(structure)
        # G0W0 calculation with reduced cutoff for the response function
        self.incar_settings.update({"ALGO": "GW0", "ENCUTGW": 250,
                                    "LWAVE": "FALSE", "NELM": 1})
        self.set_dens(spec)
        self.nomega_max = 2 * self.get_kpoints(structure).kpts[0][0]**3
        nomega = npar * int(self.nomega_max / npar)
        self.set_gw_bands(15)
        self.incar_settings.update({"NPAR": npar})
        self.incar_settings.update({"NOMEGA": nomega})
        self.tests = self.__class__.get_defaults_tests()

    def wannier_on(self):
        self.incar_settings.update({"LWANNIER90_RUN": ".TRUE."})
        self.incar_settings.update({"LWRITE_MMN_AMN": ".TRUE."})

    def spectral_off(self):
        """
        Method to switch the use of the spectral decomposition of the response function of
        this may be used to reduce memory demands if the calculation crashes due to memory shortage
        """
        self.incar_settings.update({"LSPECTRAL": ".False."})

    def gw0_on(self, niter=4, gwbandsfac=4, qpsc=False):
        """
        Method to switch to gw0 calculation with standard 4 iterations
        """
        # set the number of iterations of GW0
        self.incar_settings.update({"NELM": niter})
        # set the number of bands to update in the iteration of G
        npar = self.get_npar(self.structure)
        nbandsgw = self.get_bands(self.structure)*gwbandsfac
        nbandsgw = npar * int(nbandsgw / npar)
        self.incar_settings.update({"NBANDSGW": nbandsgw})
        # if set also updat the orbitals 'quasi particle self-consistency'
        if qpsc:
            self.incar_settings.update({"ALGO": "scGW0"})
        # todo update tests ....

    def set_prec_high(self):
        super(GWG0W0VaspInputSet, self).set_prec_high()
        self.incar_settings.update({"ENCUTGW": 400, "NOMEGA": int(self.incar_settings["NOMEGA"]*1.5)})
        self.incar_settings.update({"PRECFOCK": "accurate"})


class Wannier90InputSet():
    """
    class containing the input parameters for the wannier90.win file
    """
    def __init__(self, spec):
        self.file_name = "wannier90.win"
        self.settings = {"bands_plot": "true", "num_wann": 2, "num_bands": 4}
        self.parameters = {"n_include_bands": 1}
        self.spec = spec

    def make_kpoint_path(self, structure, f):
        f.write("\nbegin kpoint_path\n")
        line = str(structure.vbm_l) + " " + str(structure.vbm[0]) + " " + str(structure.vbm[1]) + " " + str(structure.vbm[2])
        line = line + " " + str(structure.cbm_l) + " " + str(structure.cbm[0]) + " " + str(structure.cbm[1]) + " " + str(structure.cbm[2])
        f.write(line)
        f.write("\nend kpoint_path\n\n")
        pass

    def make_exclude_bands(self, structure, f):
        nocc = GWscDFTPrepVaspInputSet(structure, self.spec).get_electrons(structure) / 2
        n1 = str(int(1))
        n2 = str(int(nocc - self.parameters["n_include_bands"]))
        n3 = str(int(nocc + 1 + self.parameters["n_include_bands"]))
        n4 = str(int(GWG0W0VaspInputSet(structure, self.spec).incar_settings["NBANDS"]))
        line = "exclude_bands : " + n1 + "-" + n2 + ", " + n3 + "-" + n4 + "\n"
        f.write(line)
        #todo there is still a bug here...
        pass

    def write_file(self, structure, path):
        f = open(os.path.join(path, self.file_name), mode='w')
        f.write("bands_plot = ")
        f.write(self.settings["bands_plot"])
        f.write("\n")
        self.make_kpoint_path(structure, f)
        f.write("num_wann  = ")
        f.write(str(self.settings["num_wann"]))
        f.write("\n")
        f.write("num_bands = ")
        f.write(str(self.settings["num_bands"]))
        f.write("\n")
        self.make_exclude_bands(structure, f)
        f.close()


class SingleVaspGWWork():
    """
    Create VASP input for a single standard G0W0 and GW0 calculation step
    the combination of job and option specifies what needs to be created
    """
    def __init__(self, structure, job, spec, option=None, converged=False):
        self.structure = structure
        self.job = job
        self.spec = spec
        self.option = option
        self.converged = converged

    def create_input(self):
        """
        create vasp input
        """
        option_name = ''
        path_add = ''
        if self.spec['converge'] and self.converged:
            path_add = '.conv'
        if self.option is None:
            path = s_name(self.structure)
        else:
            path = os.path.join(s_name(self.structure) + path_add,
                                str(self.option['test_prep'])+str(self.option['value_prep']))
            if 'test' in self.option.keys():
                option_name = '.'+str(self.option['test'])+str(self.option['value'])
        if self.job == 'prep':

            inpset = GWscDFTPrepVaspInputSet(self.structure, self.spec, functional=self.spec['functional'])
            if self.spec['converge'] and not self.converged:
                spec_tmp = self.spec.copy()
                spec_tmp.update({'kp_grid_dens': 2})
                inpset = GWscDFTPrepVaspInputSet(self.structure, spec_tmp, functional=self.spec['functional'])
                inpset.incar_settings.update({"ENCUT": 800})
            if self.spec['test'] or self.spec['converge']:
                if self.option['test_prep'] in GWscDFTPrepVaspInputSet.get_defaults_convs().keys() or self.option['test_prep'] in GWscDFTPrepVaspInputSet.get_defaults_tests().keys():
                    inpset.set_test(self.option['test_prep'], self.option['value_prep'])
            if self.spec["prec"] == "h":
                inpset.set_prec_high()
            inpset.write_input(self.structure, path)

            inpset = GWDFTDiagVaspInputSet(self.structure, self.spec, functional=self.spec['functional'])
            if self.spec["prec"] == "h":
                inpset.set_prec_high()
            if self.spec['converge'] and not self.converged:
                spec_tmp = self.spec.copy()
                spec_tmp.update({'kp_grid_dens': 2})
                inpset = GWDFTDiagVaspInputSet(self.structure, spec_tmp, functional=self.spec['functional'])
                inpset.incar_settings.update({"ENCUT": 800})
            if self.spec['test'] or self.spec['converge']:
                inpset.set_test(self.option['test_prep'], self.option['value_prep'])
            inpset.get_incar(self.structure).write_file(os.path.join(path, 'INCAR.DIAG'))

        if self.job == 'G0W0':

            inpset = GWG0W0VaspInputSet(self.structure, self.spec, functional=self.spec['functional'])
            if self.spec['converge'] and not self.converged:
                spec_tmp = self.spec.copy()
                spec_tmp.update({'kp_grid_dens': 2})
                inpset = GWG0W0VaspInputSet(self.structure, spec_tmp, functional=self.spec['functional'])
                inpset.incar_settings.update({"ENCUT": 800})
            if self.spec['test'] or self.spec['converge']:
                inpset.set_test(self.option['test_prep'], self.option['value_prep'])
                inpset.set_test(self.option['test'], self.option['value'])
            if self.spec["prec"] == "h":
                inpset.set_prec_high()
            if self.spec['kp_grid_dens'] > 20:
                #inpset.wannier_on()
                inpset.write_input(self.structure, os.path.join(path, 'G0W0'+option_name))
                #w_inpset = Wannier90InputSet(self.spec)
                #w_inpset.write_file(self.structure, os.path.join(path, 'G0W0'+option_name))
            else:
                inpset.write_input(self.structure, os.path.join(path, 'G0W0'+option_name))

        if self.job == 'GW0':

            inpset = GWG0W0VaspInputSet(self.structure, self.spec, functional=self.spec['functional'])
            if self.spec['converge'] and not self.converged:
                spec_tmp = self.spec.copy()
                spec_tmp.update({'kp_grid_dens': 2})
                inpset = GWG0W0VaspInputSet(self.structure, spec_tmp, functional=self.spec['functional'])
                inpset.incar_settings.update({"ENCUT": 800})
            if self.spec['test'] or self.spec['converge']:
                inpset.set_test(self.option['test_prep'], self.option['value_prep'])
                inpset.set_test(self.option['test'], self.option['value'])
            if self.spec["prec"] == "h":
                inpset.set_prec_high()
            inpset.gw0_on()
            if self.spec['kp_grid_dens'] > 20:
                #inpset.wannier_on()
                inpset.write_input(self.structure, os.path.join(path, 'GW0'+option_name))
                #w_inpset = Wannier90InputSet(self.spec)
                #w_inpset.write_file(self.structure, os.path.join(path, 'GW0'+option_name))
            else:
                inpset.write_input(self.structure, os.path.join(path, 'GW0'+option_name))

        if self.job == 'scGW0':

            inpset = GWG0W0VaspInputSet(self.structure, self.spec, functional=self.spec['functional'])
            if self.spec['converge'] and not self.converged:
                spec_tmp = self.spec.copy()
                spec_tmp.update({'kp_grid_dens': 2})
                inpset = GWG0W0VaspInputSet(self.structure, spec_tmp, functional=self.spec['functional'])
                inpset.incar_settings.update({"ENCUT": 800})
            if self.spec['test'] or self.spec['converge']:
                inpset.set_test(self.option['test_prep'], self.option['value_prep'])
                inpset.set_test(self.option['test'], self.option['value'])
            if self.spec["prec"] == "h":
                inpset.set_prec_high()
            inpset.gw0_on(qpsc=True)
            if self.spec['kp_grid_dens'] > 20:
                inpset.wannier_on()
                inpset.write_input(self.structure, os.path.join(path, 'scGW0'+option_name))
                w_inpset = Wannier90InputSet(self.spec)
                w_inpset.write_file(self.structure, os.path.join(path, 'scGW0'+option_name))
            else:
                inpset.write_input(self.structure, os.path.join(path, 'scGW0'+option_name))

    def create_job_script(self, add_to_collection=True, mode='pbspro'):
        if mode == 'slurm':
            """
            Create job script for ceci.
            """
            npar = GWscDFTPrepVaspInputSet(self.structure, self.spec,
                                             functional=self.spec['functional']).get_npar(self.structure)
            if self.option is not None:
                option_prep_name = str(self.option['test_prep']) + str(self.option['value_prep'])
                if 'test' in self.option.keys():
                    option_name = str('.') + str(self.option['test']) + str(self.option['value'])
            else:
                option_prep_name = option_name = ''
            # npar = int(os.environ['NPARGWCALC'])
            header = ("#!/bin/bash \n"
                      "## standard header for Ceci clusters ## \n"
                      "#SBATCH --mail-user=michiel.vansetten@uclouvain.be \n"
                      "#SBATCH --mail-type=ALL\n"
                      "#SBATCH --time=2-24:0:0 \n"
                      "#SBATCH --cpus-per-task=1 \n"
                      "#SBATCH --mem-per-cpu=4000 \n")
            path_add = ''
            if self.spec['converge'] and self.converged:
                path_add = '.conv'
            if self.job == 'prep':
                path = os.path.join(s_name(self.structure) + path_add, option_prep_name)
                # create this job
                job_file = open(name=os.path.join(path, 'job'), mode='w')
                job_file.write(header)
                job_file.write('#SBATCH --job-name='+s_name(self.structure)+self.job+'\n')
                job_file.write('#SBATCH --ntasks='+str(npar)+'\n')
                job_file.write('module load vasp \n')
                job_file.write('mpirun vasp \n')
                job_file.write('cp OUTCAR OUTCAR.sc \n')
                job_file.write('cp INCAR.DIAG INCAR \n')
                job_file.write('mpirun vasp \n')
                job_file.write('cp OUTCAR OUTCAR.diag \n')
                job_file.close()
                os.chmod(os.path.join(path, 'job'), stat.S_IRWXU)
                if add_to_collection:
                    job_file = open("job_collection", mode='a')
                    job_file.write('cd ' + path + ' \n')
                    job_file.write('sbatch job \n')
                    job_file.write('cd .. \n')
                    job_file.close()
                    os.chmod("job_collection", stat.S_IRWXU)
            if self.job in ['G0W0', 'GW0', 'scGW0']:
                path = os.path.join(s_name(self.structure) + path_add, option_prep_name, self.job + option_name)
                # create this job
                job_file = open(name=path+'/job', mode='w')
                job_file.write(header)
                job_file.write('#SBATCH --job-name='+s_name(self.structure)+self.job+'\n')
                job_file.write('#SBATCH --ntasks='+str(npar)+'\n')
                job_file.write('module load vasp/5.2_par_wannier90 \n')
                job_file.write('cp ../CHGCAR ../WAVECAR ../WAVEDER . \n')
                job_file.write('mpirun vasp \n')
                job_file.write('rm W* \n')
                #job_file.write('workon pymatgen-GW; get_gap > gap; deactivate')
                #job_file.write('echo '+path+'`get_gap` >> ../../gaps.dat')
                job_file.close()
                os.chmod(path+'/job', stat.S_IRWXU)
                path = os.path.join(s_name(self.structure) + path_add, option_prep_name)
                # 'append submission of this job script to that of prep for this structure'
                if add_to_collection:
                    job_file = open(name=os.path.join(path, 'job'), mode='a')
                    job_file.write('cd ' + self.job + option_name + ' \n')
                    job_file.write('sbatch job \n')
                    job_file.write('cd .. \n')
                    job_file.close()
        elif mode == 'pbspro':
            """
            Create job script for pbse pro Zenobe.
            """
            npar = GWscDFTPrepVaspInputSet(self.structure, self.spec,
                                           functional=self.spec['functional']).get_npar(self.structure)
            #npar = 96
            if self.option is not None:
                option_prep_name = str(self.option['test_prep']) + str(self.option['value_prep'])
                if 'test' in self.option.keys():
                    option_name = str('.') + str(self.option['test']) + str(self.option['value'])
            else:
                option_prep_name = option_name = ''
            # npar = int(os.environ['NPARGWCALC'])
            header = str("#!/bin/bash \n" +
                         "## standard header for zenobe ## \n" +
                         "#!/bin/bash \n" +
                         "#PBS -q main\n" +
                         "#PBS -l walltime=24:0:00\n" +
                         "#PBS -r y \n" +
                         "#PBS -m abe\n" +
                         "#PBS -M michiel.vansetten@uclouvain.be\n" +
                         "#PBS -W group_list=naps\n" +
                         "#PBS -l pvmem=1900mb\n")
            path_add = ''
            if self.spec['converge'] and self.converged:
                path_add = '.conv'
            if self.job == 'prep':
                path = os.path.join(s_name(self.structure) + path_add, option_prep_name)
                abs_path = os.path.abspath(path)
                # create this job
                job_file = open(name=os.path.join(path, 'job'), mode='w')
                job_file.write(header)
                job_file.write("#PBS -l select=%s:ncpus=1:vmem=1900mb:mpiprocs=1:ompthreads=1\n" % str(npar))
                job_file.write('#PBS -o %s/queue.qout\n#PBS -e %s/queue.qerr\ncd %s\n' % (abs_path, abs_path, abs_path))
                job_file.write('mpirun -n %s vasp \n' % str(npar))
                job_file.write('cp OUTCAR OUTCAR.sc \n')
                job_file.write('cp INCAR.DIAG INCAR \n')
                job_file.write('mpirun -n %s vasp \n' % str(npar))
                job_file.write('cp OUTCAR OUTCAR.diag \n')
                job_file.close()
                os.chmod(os.path.join(path, 'job'), stat.S_IRWXU)
                if add_to_collection:
                    job_file = open("job_collection", mode='a')
                    job_file.write('cd ' + path + ' \n')
                    job_file.write('qsub job \n')
                    job_file.write('cd ../.. \n')
                    job_file.close()
                    os.chmod("job_collection", stat.S_IRWXU)
            if self.job in ['G0W0', 'GW0', 'scGW0']:
                path = os.path.join(s_name(self.structure) + path_add, option_prep_name, self.job + option_name)
                abs_path = os.path.abspath(path)
                # create this job
                job_file = open(name=path+'/job', mode='w')
                job_file.write(header)
                job_file.write("#PBS -l select=%s:ncpus=1:vmem=1000mb:mpiprocs=1:ompthreads=1\n" % str(npar))
                job_file.write('#PBS -o %s/queue.qout\n#PBS -e %s/queue.qerr\ncd %s\n' % (abs_path, abs_path, abs_path))
                job_file.write('cp ../CHGCAR ../WAVECAR ../WAVEDER . \n')
                job_file.write('mpirun -n %s vasp \n' % str(npar))
                job_file.write('rm W* \n')
                #job_file.write('workon pymatgen-GW; get_gap > gap; deactivate')
                #job_file.write('echo '+path+'`get_gap` >> ../../gaps.dat')
                job_file.close()
                os.chmod(path+'/job', stat.S_IRWXU)
                path = os.path.join(s_name(self.structure) + path_add, option_prep_name)
                # 'append submission of this job script to that of prep for this structure'
                if add_to_collection:
                    job_file = open(name=os.path.join(path, 'job'), mode='a')
                    job_file.write('cd ' + self.job + option_name + ' \n')
                    job_file.write('qsub job \n')
                    job_file.write('cd .. \n')
                    job_file.close()



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import unicode_literals

"""
Common test support for pymatgen test scripts.

This single module should provide all the common functionality for pymatgen
tests in a single location, so that test scripts can just import it and work
right away.
"""

import unittest2 as unittest
import tempfile
import numpy.testing.utils as nptu
from six.moves import zip
from io import open
import os
import json

from monty.json import MontyDecoder
from monty.serialization import loadfn


from monty.json import MSONable

class PymatgenTest(unittest.TestCase):
    """
    Extends unittest.TestCase with functions (taken from numpy.testing.utils)
    that support the comparison of arrays.
    """
    MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
    STRUCTURES_DIR = os.path.join(MODULE_DIR, "structures")

    """
    Dict for test structures to aid testing.
    """
    TEST_STRUCTURES = {}
    for fn in os.listdir(STRUCTURES_DIR):
        TEST_STRUCTURES[fn.rsplit(".", 1)[0]] = loadfn(os.path.join(
            STRUCTURES_DIR, fn), cls=MontyDecoder)

    @classmethod
    def get_structure(cls, name):
        return cls.TEST_STRUCTURES[name].copy()

    @staticmethod
    def assert_almost_equal(actual, desired, decimal=7, err_msg='',
                            verbose=True):
        """
        Alternative naming for assertArrayAlmostEqual.
        """
        return PymatgenTest.assertArrayAlmostEqual(
            actual, desired, decimal, err_msg, verbose)

    @staticmethod
    def assert_equal(actual, desired, err_msg='', verbose=True):
        """
        Alternative naming for assertArrayEqual.
        """
        return PymatgenTest.assertArrayEqual(actual, desired,
                                             err_msg=err_msg, verbose=verbose)

    @staticmethod
    def assertArrayAlmostEqual(actual, desired, decimal=7, err_msg='',
                               verbose=True):
        """
        Tests if two arrays are almost equal to a tolerance. The CamelCase
        naming is so that it is consistent with standard unittest methods.
        """
        return nptu.assert_almost_equal(actual, desired, decimal, err_msg,
                                        verbose)

    @staticmethod
    def assertArrayEqual(actual, desired, err_msg='', verbose=True):
        """
        Tests if two arrays are equal. The CamelCase naming is so that it is
         consistent with standard unittest methods.
        """
        return nptu.assert_equal(actual, desired, err_msg=err_msg,
                                 verbose=verbose)

    def serialize_with_pickle(self, objects, protocols=None, test_eq=True):
        """
        Test whether the object(s) can be serialized and deserialized with pickle.
        This method tries to serialize the objects with pickle and the protocols
        specified in input. Then it deserializes the pickle format and compares
        the two objects with the __eq__ operator if test_eq == True.

        Args:
            objects: Object or list of objects.
            protocols: List of pickle protocols to test. If protocols is None, HIGHEST_PROTOCOL is tested.

        Returns:
            Nested list with the objects deserialized with the specified protocols.
        """
        # Use the python version so that we get the traceback in case of errors
        import pickle as pickle
        from pymatgen.serializers.pickle_coders import pmg_pickle_load, pmg_pickle_dump

        # Build a list even when we receive a single object.
        got_single_object = False
        if not isinstance(objects, (list, tuple)):
            got_single_object = True
            objects = [objects]

        if protocols is None:
            #protocols = set([0, 1, 2] + [pickle.HIGHEST_PROTOCOL])
            protocols = [pickle.HIGHEST_PROTOCOL]

        # This list will contains the object deserialized with the different protocols.
        objects_by_protocol, errors = [], []

        for protocol in protocols:
            # Serialize and deserialize the object.
            mode = "wb"
            fd, tmpfile = tempfile.mkstemp(text="b" not in mode)

            try:
                with open(tmpfile, mode) as fh:
                    #pickle.dump(objects, fh, protocol=protocol)
                    pmg_pickle_dump(objects, fh, protocol=protocol)
            except Exception as exc:
                errors.append("pickle.dump with protocol %s raised:\n%s" % (protocol, str(exc)))
                continue

            try:
                with open(tmpfile, "rb") as fh:
                    #new_objects = pickle.load(fh)
                    new_objects = pmg_pickle_load(fh)
            except Exception as exc:
                errors.append("pickle.load with protocol %s raised:\n%s" % (protocol, str(exc)))
                continue

            # Test for equality
            if test_eq:
                for old_obj, new_obj in zip(objects, new_objects):
                    self.assert_equal(old_obj, new_obj)

            # Save the deserialized objects and test for equality.
            objects_by_protocol.append(new_objects)

        if errors:
            raise ValueError("\n".join(errors))

        # Return nested list so that client code can perform additional tests.
        if got_single_object:
            return [o[0] for o in objects_by_protocol]
        else:
            return objects_by_protocol

    def tmpfile_write(self, string):
        """
        Write string to a temporary file. Returns the name of the temporary file.
        """
        fd, tmpfile = tempfile.mkstemp(text=True)

        with open(tmpfile, "w") as fh:
            fh.write(string)

        return tmpfile

    def assertMSONable(self, obj):
        """
        Tests if obj is MSONable and tries to verify whether the contract is fullfilled.
        """
        self.assertIsInstance(obj, MSONable)
        self.assertDictEqual(obj.as_dict(), obj.__class__.from_dict(obj.as_dict()).as_dict())
        json.loads(obj.to_json(), cls=MontyDecoder)



#!/usr/bin/env python


__author__ = 'waroquiers'

import unittest2 as unittest
from pymatgen.analysis.chemenv.utils.math_utils import prime_factors, divisors, get_center_of_arc, get_linearly_independent_vectors
from pymatgen.analysis.chemenv.utils.math_utils import scale_and_clamp, smoothstep, smootherstep, cosinus_step, power3_step
from pymatgen.analysis.chemenv.utils.math_utils import powern_parts_step
from pymatgen.analysis.chemenv.utils.math_utils import _cartesian_product
import numpy as np


class MathUtilsTest(unittest.TestCase):

    def test_list_cartesian_product(self):
        list_of_lists = [[0, 1], [2, 5, 4], [5]]
        self.assertEqual(_cartesian_product(lists=list_of_lists),
                         [[0, 2, 5], [1, 2, 5], [0, 5, 5], [1, 5, 5], [0, 4, 5], [1, 4, 5]])
        list_of_lists = [[0, 1], [2, 5, 4], []]
        self.assertEqual(_cartesian_product(lists=list_of_lists),
                         [])
        list_of_lists = [[1], [3], [2]]
        self.assertEqual(_cartesian_product(lists=list_of_lists),
                         [[1, 3, 2]])
        list_of_lists = [[7]]
        self.assertEqual(_cartesian_product(lists=list_of_lists),
                         [[7]])

    def test_math_utils(self):
        ff = prime_factors(250)
        self.assertEqual(ff, [5, 5, 5, 2])
        div = divisors(560)
        self.assertEqual(div, [1, 2, 4, 5, 7, 8, 10, 14, 16, 20, 28, 35, 40, 56, 70, 80, 112, 140, 280, 560])
        center = get_center_of_arc([0.0, 0.0], [1.0, 0.0], 0.5)
        self.assertEqual(center, (0.5, 0.0))

    def test_linearly_independent_vectors(self):
        v1 = np.array([1, 0, 0])
        v2 = np.array([0, 1, 0])
        v3 = np.array([0, 0, 1])
        v4 = np.array([-1, 0, 0])
        v5 = np.array([1, 1, 0])
        independent_vectors = get_linearly_independent_vectors([v1, v2, v3])
        self.assertEqual(len(independent_vectors), 3)
        independent_vectors = get_linearly_independent_vectors([v1, v2, v4])
        self.assertEqual(len(independent_vectors), 2)
        independent_vectors = get_linearly_independent_vectors([v1, v2, v5])
        self.assertEqual(len(independent_vectors), 2)
        independent_vectors = get_linearly_independent_vectors([v1, v2, v3, v4, v5])
        self.assertEqual(len(independent_vectors), 3)

    def test_scale_and_clamp(self):
        edge0 = 7.0
        edge1 = 11.0
        clamp0 = 0.0
        clamp1 = 1.0
        vals = np.linspace(5.0, 12.0, num=8)
        self.assertEqual(scale_and_clamp(vals, edge0, edge1, clamp0, clamp1).tolist(),
                         [0.0, 0.0, 0.0, 0.25, 0.5, 0.75, 1.0, 1.0])

    def test_smoothstep(self):
        vals = np.linspace(5.0, 12.0, num=8)
        self.assertEqual(smoothstep(vals, edges=[0.0, 1.0]).tolist(), [1.0]*8)
        self.assertEqual(smoothstep(vals, edges=[7.0, 11.0]).tolist(),
                         [0.0, 0.0, 0.0, 0.15625, 0.5, 0.84375, 1.0, 1.0])

    def test_smootherstep(self):
        vals = np.linspace(5.0, 12.0, num=8)
        self.assertEqual(smootherstep(vals, edges=[0.0, 1.0]).tolist(), [1.0]*8)
        self.assertEqual(smootherstep(vals, edges=[7.0, 11.0]).tolist(),
                         [0.0, 0.0, 0.0, 0.103515625, 0.5, 0.896484375, 1.0, 1.0])

    def test_power3_step(self):
        vals = np.linspace(5.0, 12.0, num=8)
        self.assertEqual(power3_step(vals, edges=[0.0, 1.0]).tolist(), [1.0]*8)
        self.assertEqual(power3_step(vals, edges=[7.0, 11.0]).tolist(),
                         [0.0, 0.0, 0.0, 0.15625, 0.5, 0.84375, 1.0, 1.0])

    def test_cosinus_step(self):
        vals = np.linspace(5.0, 12.0, num=8)
        self.assertEqual(cosinus_step(vals, edges=[0.0, 1.0]).tolist(), [1.0]*8)
        self.assertEqual(cosinus_step(vals, edges=[7.0, 11.0]).tolist(),
                         [0.0, 0.0, 0.0, 0.14644660940672627, 0.5, 0.8535533905932737, 1.0, 1.0])

    def test_powern_parts_step(self):
        vals = np.linspace(5.0, 12.0, num=8)
        self.assertEqual(powern_parts_step(vals, edges=[0.0, 1.0], nn=2).tolist(), [1.0]*8)
        self.assertEqual(powern_parts_step(vals, edges=[0.0, 1.0], nn=3).tolist(), [1.0]*8)
        self.assertEqual(powern_parts_step(vals, edges=[0.0, 1.0], nn=4).tolist(), [1.0]*8)
        self.assertEqual(powern_parts_step(vals, edges=[7.0, 11.0], nn=2).tolist(),
                         [0.0, 0.0, 0.0, 0.125, 0.5, 0.875, 1.0, 1.0])
        self.assertEqual(powern_parts_step(vals, edges=[7.0, 11.0], nn=3).tolist(),
                         [0.0, 0.0, 0.0, 0.0625, 0.5, 0.9375, 1.0, 1.0])
        self.assertEqual(powern_parts_step(vals, edges=[7.0, 11.0], nn=4).tolist(),
                         [0.0, 0.0, 0.0, 0.03125, 0.5, 0.96875, 1.0, 1.0])

if __name__ == "__main__":
    unittest.main()